@@ -107,6 +107,44 @@ double calc_kurtosis(size_t ix_arr[], size_t st, size_t end, real_t x[], Missing
     }
 }
 
+/* TODO: is this algorithm correct? */
+template <class real_t, class mapping>
+double calc_kurtosis_weighted(size_t ix_arr[], size_t st, size_t end, real_t x[],
+                              MissingAction missing_action, mapping w)
+{
+    long double m = 0;
+    long double M2 = 0, M3 = 0, M4 = 0;
+    long double delta, delta_s, delta_div;
+    long double diff;
+    long double n = 0;
+    long double out;
+    long double n_prev = 0.;
+    double w_this;
+
+    for (size_t row = st; row <= end; row++)
+    {
+        if (!is_na_or_inf(x[ix_arr[row]]))
+        {
+            w_this = w[ix_arr[row]];
+            n += w_this;
+
+            delta      =  x[ix_arr[row]] - m;
+            delta_div  =  delta / n;
+            delta_s    =  delta_div * delta_div;
+            diff       =  delta * (delta_div * n_prev);
+            n_prev   =  n;
+
+            m   +=  w_this * (delta_div);
+            M4  +=  w_this * (diff * delta_s * (n * n - 3 * n + 3) + 6 * delta_s * M2 - 4 * delta_div * M3);
+            M3  +=  w_this * (diff * delta_div * (n - 2) - 3 * delta_div * M2);
+            M2  +=  w_this * (diff);
+        }
+    }
+
+    out = ( M4 / M2 ) * ( n / M2 );
+    return (!is_na_or_inf(out) && out > 0.)? out : 0.;
+}
+
 
 /* TODO: make these compensated sums */
 /* TODO: can this use the same algorithm as above but with a correction at the end,
@@ -206,6 +244,106 @@ double calc_kurtosis(size_t ix_arr[], size_t st, size_t end, size_t col_num,
 }
 
 
+template <class real_t, class sparse_ix, class mapping>
+double calc_kurtosis_weighted(size_t ix_arr[], size_t st, size_t end, size_t col_num,
+                              real_t Xc[], sparse_ix Xc_ind[], sparse_ix Xc_indptr[],
+                              MissingAction missing_action, mapping w)
+{
+    /* ix_arr must be already sorted beforehand */
+    if (Xc_indptr[col_num] == Xc_indptr[col_num + 1])
+        return 0;
+
+    long double s1 = 0;
+    long double s2 = 0;
+    long double s3 = 0;
+    long double s4 = 0;
+    double w_this;
+    long double cnt = 0.;
+    for (size_t row = st; row <= end; row++)
+        cnt += w[ix_arr[row]];
+
+    if (cnt <= 1) return 0;
+    
+    size_t st_col  = Xc_indptr[col_num];
+    size_t end_col = Xc_indptr[col_num + 1] - 1;
+    size_t curr_pos = st_col;
+    size_t ind_end_col = Xc_ind[end_col];
+    size_t *ptr_st = std::lower_bound(ix_arr + st, ix_arr + end + 1, Xc_ind[st_col]);
+
+    if (missing_action != Fail)
+    {
+        for (size_t *row = ptr_st;
+             row != ix_arr + end + 1 && curr_pos != end_col + 1 && ind_end_col >= *row;
+            )
+        {
+            if (Xc_ind[curr_pos] == *row)
+            {
+                w_this = w[*row];
+
+                if (is_na_or_inf(Xc[curr_pos]))
+                {
+                    cnt -= w_this;
+                }
+
+                else
+                {
+                    s1 += w_this * pw1(Xc[curr_pos]);
+                    s2 += w_this * pw2(Xc[curr_pos]);
+                    s3 += w_this * pw3(Xc[curr_pos]);
+                    s4 += w_this * pw4(Xc[curr_pos]);
+                }
+
+                if (row == ix_arr + end || curr_pos == end_col) break;
+                curr_pos = std::lower_bound(Xc_ind + curr_pos + 1, Xc_ind + end_col + 1, *(++row)) - Xc_ind;
+            }
+
+            else
+            {
+                if (Xc_ind[curr_pos] > *row)
+                    row = std::lower_bound(row + 1, ix_arr + end + 1, Xc_ind[curr_pos]);
+                else
+                    curr_pos = std::lower_bound(Xc_ind + curr_pos + 1, Xc_ind + end_col + 1, *row) - Xc_ind;
+            }
+        }
+    }
+
+    else
+    {
+        for (size_t *row = ptr_st;
+             row != ix_arr + end + 1 && curr_pos != end_col + 1 && ind_end_col >= *row;
+            )
+        {
+            if (Xc_ind[curr_pos] == *row)
+            {
+                s1 += pw1(Xc[curr_pos]);
+                s2 += pw2(Xc[curr_pos]);
+                s3 += pw3(Xc[curr_pos]);
+                s4 += pw4(Xc[curr_pos]);
+
+                if (row == ix_arr + end || curr_pos == end_col) break;
+                curr_pos = std::lower_bound(Xc_ind + curr_pos + 1, Xc_ind + end_col + 1, *(++row)) - Xc_ind;
+            }
+
+            else
+            {
+                if (Xc_ind[curr_pos] > *row)
+                    row = std::lower_bound(row + 1, ix_arr + end + 1, Xc_ind[curr_pos]);
+                else
+                    curr_pos = std::lower_bound(Xc_ind + curr_pos + 1, Xc_ind + end_col + 1, *row) - Xc_ind;
+            }
+        }
+    }
+
+    if (cnt <= 1 || s2 == 0 || s2 == pw2(s1)) return 0;
+    long double sn = s1 / cnt;
+    long double v  = s2 / cnt - pw2(sn);
+    if (v <= 0) return 0.;
+    long double out =  (s4 - 4 * s3 * sn + 6 * s2 * pw2(sn) - 4 * s1 * pw3(sn) + cnt * pw4(sn)) / (cnt * pw2(v));
+    return (!is_na_or_inf(out) && out > 0.)? out : 0.;
+}
+
+
+
 double calc_kurtosis(size_t ix_arr[], size_t st, size_t end, int x[], int ncat, size_t buffer_cnt[], double buffer_prob[],
                      MissingAction missing_action, CategSplit cat_split_type, RNG_engine &rnd_generator)
 {
@@ -300,6 +438,86 @@ double calc_kurtosis(size_t ix_arr[], size_t st, size_t end, int x[], int ncat,
 }
 
 
+/* TODO: this one should get a buffer preallocated from outside */
+template <class mapping>
+double calc_kurtosis_weighted(size_t ix_arr[], size_t st, size_t end, int x[], int ncat, double buffer_prob[],
+                              MissingAction missing_action, CategSplit cat_split_type, RNG_engine &rnd_generator, mapping w)
+{
+    long double cnt = 0.;
+    std::vector<long double> buffer_cnt(ncat+1, 0.);
+    double sum_kurt = 0;
+    double w_this;
+
+    for (size_t row = st; row <= end; row++)
+    {
+        w_this = w[ix_arr[row]];
+        if (x[ix_arr[row]] >= 0)
+            buffer_cnt[x[ix_arr[row]]] += w_this;
+        else
+            buffer_cnt[ncat] += w_this;
+    }
+    cnt = std::accumulate(buffer_cnt.begin(), buffer_cnt.end(), (long double)0.);
+
+    cnt -= buffer_cnt[ncat];
+    if (cnt <= 1) return 0;
+    for (int cat = 0; cat < ncat; cat++)
+        buffer_prob[cat] = buffer_cnt[cat] / cnt;
+
+    switch(cat_split_type)
+    {
+        case SubSet:
+        {
+            long double temp_v;
+            long double s1, s2, s3, s4;
+            long double coef;
+            std::uniform_real_distribution<double> runif(0, 1);
+            size_t ntry = 50;
+            for (size_t iternum = 0; iternum < 50; iternum++)
+            {
+                s1 = 0; s2 = 0; s3 = 0; s4 = 0;
+                for (int cat = 0; cat < ncat; cat++)
+                {
+                    coef = runif(rnd_generator);
+                    s1 += buffer_prob[cat] * pw1(coef);
+                    s2 += buffer_prob[cat] * pw2(coef);
+                    s3 += buffer_prob[cat] * pw3(coef);
+                    s4 += buffer_prob[cat] * pw4(coef);
+                }
+                temp_v = s2 - pw2(s1);
+                if (temp_v <= 0)
+                    ntry--;
+                else
+                    sum_kurt += (s4 - 4 * s3 * pw1(s1) + 6 * s2 * pw2(s1) - 4 * s1 * pw3(s1) + pw4(s1)) / pw2(temp_v);
+            }
+            if (!ntry)
+                return 0;
+            else
+                return ((!is_na_or_inf(sum_kurt) && sum_kurt > 0.)? sum_kurt : 0.) / (long double)ntry;
+        }
+
+        case SingleCateg:
+        {
+            double p;
+            int ncat_present = ncat;
+            for (int cat = 0; cat < ncat; cat++)
+            {
+                p = buffer_prob[cat];
+                if (p == 0)
+                    ncat_present--;
+                else
+                    sum_kurt += (p - 4 * p * pw1(p) + 6 * p * pw2(p) - 4 * p * pw3(p) + pw4(p)) / pw2(p - pw2(p));
+            }
+            if (ncat_present <= 1)
+                return 0;
+            else
+                return ((!is_na_or_inf(sum_kurt) && sum_kurt > 0.)? sum_kurt : 0.) / (double) ncat_present;
+        }
+    }
+
+    return -1; /* this will never be reached, but CRAN complains otherwise */
+}
+
+
 double expected_sd_cat(double p[], size_t n, size_t pos[])
 {
     if (n <= 1) return 0;
@@ -314,11 +532,12 @@ double expected_sd_cat(double p[], size_t n, size_t pos[])
     return std::sqrt(std::fmax(cum_var, 0.));
 }
 
-double expected_sd_cat(size_t counts[], double p[], size_t n, size_t pos[])
+template <class number>
+double expected_sd_cat(number counts[], double p[], size_t n, size_t pos[])
 {
     if (n <= 1) return 0;
 
-    size_t tot = std::accumulate(pos, pos + n, (size_t)0, [&counts](size_t tot, const size_t ix){return tot + counts[ix];});
+    number tot = std::accumulate(pos, pos + n, (number)0, [&counts](number tot, const size_t ix){return tot + counts[ix];});
     long double cnt_div = (long double) tot;
     for (size_t cat = 0; cat < n; cat++)
         p[pos[cat]] = (long double)counts[pos[cat]] / cnt_div;
@@ -326,7 +545,8 @@ double expected_sd_cat(size_t counts[], double p[], size_t n, size_t pos[])
     return expected_sd_cat(p, n, pos);
 }
 
-double expected_sd_cat_single(size_t counts[], double p[], size_t n, size_t pos[], size_t cat_exclude, size_t cnt)
+template <class number>
+double expected_sd_cat_single(number counts[], double p[], size_t n, size_t pos[], size_t cat_exclude, number cnt)
 {
     if (cat_exclude == 0)
         return expected_sd_cat(counts, p, n-1, pos + 1);
@@ -362,7 +582,8 @@ double expected_sd_cat_single(size_t counts[], double p[], size_t n, size_t pos[
 /* Note: this isn't exactly comparable to the pooled gain from numeric variables,
    but among all the possible options, this is what happens to end up in the most
    similar scale when considering standardized gain. */
-double categ_gain(size_t cnt_left, size_t cnt_right,
+template <class number>
+double categ_gain(number cnt_left, number cnt_right,
                   long double s_left, long double s_right,
                   long double base_info, long double cnt)
 {
@@ -453,6 +674,10 @@ double categ_gain(size_t cnt_left, size_t cnt_right,
     (1. - (1./(sd))*(  ( ((real_t)(cnt_left))/(cnt) )*(sd_left) + ( ((real_t)(cnt_right)/(cnt)) )*(sd_right)  ))
 
 
+/* TODO: these functions would not take into account observation weights if available.
+   Need to create a 'weighted' version of each. */
+
+
 /* TODO: make this a compensated sum */
 template <class real_t, class real_t_>
 double find_split_rel_gain_t(real_t_ *restrict x, size_t n, double &split_point)
@@ -549,6 +774,32 @@ real_t calc_sd_right_to_left(real_t_ *restrict x, size_t n, double *restrict sd_
     return std::sqrt(running_ssq / (real_t)n);
 }
 
+template <class real_t_>
+long double calc_sd_right_to_left_weighted(real_t_ *restrict x, size_t n, double *restrict sd_arr,
+                                           double *restrict w, long double &cumw, size_t *restrict sorted_ix)
+{
+    long double running_mean = 0;
+    long double running_ssq = 0;
+    long double mean_prev = x[sorted_ix[n-1]];
+    long double cnt = 0;
+    double w_this;
+    for (size_t row = 0; row < n-1; row++)
+    {
+        w_this = w[sorted_ix[n-row-1]];
+        cnt += w_this;
+        running_mean   += (x[sorted_ix[n-row-1]] - running_mean) / cnt;
+        running_ssq    += w_this * ((x[sorted_ix[n-row-1]] - running_mean) * (x[sorted_ix[n-row-1]] - mean_prev));
+        mean_prev       =  running_mean;
+        sd_arr[n-row-1] = (row == 0)? 0. : std::sqrt(running_ssq / cnt);
+    }
+    w_this = w[sorted_ix[0]];
+    cnt += w_this;
+    running_mean   += (x[sorted_ix[0]] - running_mean) / cnt;
+    running_ssq    += w_this * ((x[sorted_ix[0]] - running_mean) * (x[sorted_ix[0]] - mean_prev));
+    cumw = cnt;
+    return std::sqrt(running_ssq / cnt);
+}
+
 template <class real_t, class real_t_>
 real_t calc_sd_right_to_left(real_t_ *restrict x, real_t_ xmean, size_t ix_arr[], size_t st, size_t end, double *restrict sd_arr)
 {
@@ -568,6 +819,33 @@ real_t calc_sd_right_to_left(real_t_ *restrict x, real_t_ xmean, size_t ix_arr[]
     return std::sqrt(running_ssq / (real_t)n);
 }
 
+template <class real_t_, class mapping>
+long double calc_sd_right_to_left_weighted(real_t_ *restrict x, real_t_ xmean, size_t ix_arr[], size_t st, size_t end,
+                                           double *restrict sd_arr, mapping w, long double &cumw)
+{
+    long double running_mean = 0;
+    long double running_ssq = 0;
+    real_t_ mean_prev = x[ix_arr[end]] - xmean;
+    size_t n = end - st + 1;
+    long double cnt = 0;
+    double w_this;
+    for (size_t row = 0; row < n-1; row++)
+    {
+        w_this = w[ix_arr[end-row]];
+        cnt += w_this;
+        running_mean   += ((x[ix_arr[end-row]] - xmean) - running_mean) / cnt;
+        running_ssq    += w_this * (((x[ix_arr[end-row]] - xmean) - running_mean) * ((x[ix_arr[end-row]] - xmean) - mean_prev));
+        mean_prev       =  running_mean;
+        sd_arr[n-row-1] = (row == 0)? 0. : std::sqrt(running_ssq / cnt);
+    }
+    w_this = w[ix_arr[st]];
+    cnt += w_this;
+    running_mean   += ((x[ix_arr[st]] - xmean) - running_mean) / cnt;
+    running_ssq    += w_this * (((x[ix_arr[st]] - xmean) - running_mean) * ((x[ix_arr[st]] - xmean) - mean_prev));
+    cumw = cnt;
+    return std::sqrt(running_ssq / cnt);
+}
+
 template <class real_t, class real_t_>
 double find_split_std_gain_t(real_t_ *restrict x, size_t n, double *restrict sd_arr,
                              GainCriterion criterion, double min_gain, double &split_point)
@@ -611,6 +889,45 @@ double find_split_std_gain(real_t_ *restrict x, size_t n, double *restrict sd_ar
         return find_split_std_gain_t<long double, real_t_>(x, n, sd_arr, criterion, min_gain, split_point);
 }
 
+template <class real_t>
+double find_split_std_gain_weighted(real_t *restrict x, size_t n, double *restrict sd_arr,
+                                    GainCriterion criterion, double min_gain, double &split_point,
+                                    double *restrict w, size_t *restrict sorted_ix)
+{
+    long double cumw;
+    double full_sd = calc_sd_right_to_left_weighted(x, n, sd_arr, w, cumw, sorted_ix);
+    long double running_mean = 0;
+    long double running_ssq = 0;
+    long double mean_prev = x[sorted_ix[0]];
+    double best_gain = -HUGE_VAL;
+    double this_sd, this_gain;
+    double w_this;
+    long double currw = 0;
+
+    for (size_t row = 0; row < n-1; row++)
+    {
+        w_this = w[row];
+        currw += w_this;
+        running_mean   += (x[sorted_ix[row]] - running_mean) / currw;
+        running_ssq    += w_this * ((x[sorted_ix[row]] - running_mean) * (x[sorted_ix[row]] - mean_prev));
+        mean_prev       =  running_mean;
+        if (x[sorted_ix[row]] == x[sorted_ix[row+1]])
+            continue;
+
+        this_sd = (row == 0)? 0. : std::sqrt(running_ssq / currw);
+        this_gain = (criterion == Pooled)?
+                    pooled_gain(full_sd, cumw, this_sd, sd_arr[row+1], currw, cumw-currw)
+                        :
+                    sd_gain(full_sd, this_sd, sd_arr[row+1]);
+        if (this_gain > best_gain && this_gain > min_gain)
+        {
+            best_gain = this_gain;
+            split_point = avg_between(x[sorted_ix[row]], x[sorted_ix[row+1]]);
+        }
+    }
+    return best_gain;
+}
+
 template <class real_t, class real_t_>
 double find_split_std_gain_t(real_t_ *restrict x, real_t_ xmean, size_t ix_arr[], size_t st, size_t end, double *restrict sd_arr,
                              GainCriterion criterion, double min_gain, double &split_point, size_t &split_ix)
@@ -655,6 +972,44 @@ double find_split_std_gain(real_t_ *restrict x, real_t_ xmean, size_t ix_arr[],
         return find_split_std_gain_t<long double, real_t_>(x, xmean, ix_arr, st, end, sd_arr, criterion, min_gain, split_point, split_ix);
 }
 
+template <class real_t, class mapping>
+double find_split_std_gain_weighted(real_t *restrict x, real_t xmean, size_t ix_arr[], size_t st, size_t end, double *restrict sd_arr,
+                                    GainCriterion criterion, double min_gain, double &split_point, size_t &split_ix, mapping w)
+{
+    long double cumw;
+    double full_sd = calc_sd_right_to_left_weighted(x, xmean, ix_arr, st, end, sd_arr, w, cumw);
+    long double running_mean = 0;
+    long double running_ssq = 0;
+    long double mean_prev = x[ix_arr[st]] - xmean;
+    double best_gain = -HUGE_VAL;
+    long double currw = 0;
+    double this_sd, this_gain;
+    double w_this;
+    for (size_t row = st; row < end; row++)
+    {
+        w_this = w[ix_arr[row]];
+        currw += w_this;
+        running_mean   += ((x[ix_arr[row]] - xmean) - running_mean) / currw;
+        running_ssq    += w_this * (((x[ix_arr[row]] - xmean) - running_mean) * ((x[ix_arr[row]] - xmean) - mean_prev));
+        mean_prev       =  running_mean;
+        if (x[ix_arr[row]] == x[ix_arr[row+1]])
+            continue;
+
+        this_sd = (row == st)? 0. : std::sqrt(running_ssq / currw);
+        this_gain = (criterion == Pooled)?
+                    pooled_gain(full_sd, cumw, this_sd, sd_arr[row-st+1], currw, cumw-currw)
+                        :
+                    sd_gain(full_sd, this_sd, sd_arr[row-st+1]);
+        if (this_gain > best_gain && this_gain > min_gain)
+        {
+            best_gain = this_gain;
+            split_point = avg_between(x[ix_arr[row]], x[ix_arr[row+1]]);
+            split_ix = row;
+        }
+    }
+    return best_gain;
+}
+
 
 /* for split-criterion in hyperplanes (see below for version aimed at single-variable splits) */
 double eval_guided_crit(double *restrict x, size_t n, GainCriterion criterion,
@@ -690,6 +1045,39 @@ double eval_guided_crit(double *restrict x, size_t n, GainCriterion criterion,
     return std::fmax(0., gain);
 }
 
+double eval_guided_crit_weighted(double *restrict x, size_t n, GainCriterion criterion,
+                                 double min_gain, bool as_relative_gain, double *restrict buffer_sd,
+                                 double &split_point, double &xmin, double &xmax,
+                                 double *restrict w, size_t *restrict buffer_indices)
+{
+    /* Note: the input 'x' is supposed to be a linear combination of standardized variables, so
+       all numbers are assumed to be small and in the same scale */
+    double gain;
+
+    /* here it's assumed the 'x' vector matches exactly with 'ix_arr' + 'st' */
+    if (n == 2)
+    {
+        if (x[0] == x[1]) return -HUGE_VAL;
+        split_point = avg_between(x[0], x[1]);
+        gain        = 1.;
+        if (gain > min_gain)
+            return gain;
+        else
+            return 0.;
+    }
+
+    /* sort in ascending order */
+    std::iota(buffer_indices, buffer_indices + n, (size_t)0);
+    std::sort(buffer_indices, buffer_indices + n,
+              [&x](const size_t a, const size_t b){return x[a] < x[b];});
+    xmin = x[buffer_indices[0]]; xmax = x[buffer_indices[n-1]];
+    if (x[0] == x[n-1]) return -HUGE_VAL;
+
+    gain = find_split_std_gain_weighted(x, n, buffer_sd, criterion, min_gain, split_point, w, buffer_indices);
+    /* Note: a gain of -Inf signals that the data is unsplittable. Zero signals it's below the minimum. */
+    return std::fmax(0., gain);
+}
+
 /* for split-criterion in single-variable splits */
 template <class real_t_>
 double eval_guided_crit(size_t *restrict ix_arr, size_t st, size_t end, real_t_ *restrict x,
@@ -739,6 +1127,56 @@ double eval_guided_crit(size_t *restrict ix_arr, size_t st, size_t end, real_t_
     return std::fmax(0., gain);
 }
 
+template <class real_t_, class mapping>
+double eval_guided_crit_weighted(size_t *restrict ix_arr, size_t st, size_t end, real_t_ *restrict x,
+                                 double *restrict buffer_sd, bool as_relative_gain,
+                                 size_t &split_ix, double &split_point, double &xmin, double &xmax,
+                                 GainCriterion criterion, double min_gain, MissingAction missing_action,
+                                 mapping w)
+{
+    double gain;
+
+    /* move NAs to the front if there's any, exclude them from calculations */
+    if (missing_action != Fail)
+        st = move_NAs_to_front(ix_arr, st, end, x);
+
+    if (st >= end) return -HUGE_VAL;
+    else if (st == (end-1))
+    {
+        if (x[ix_arr[st]] == x[ix_arr[end]])
+            return -HUGE_VAL;
+        split_point = avg_between(x[ix_arr[st]], x[ix_arr[end]]);
+        split_ix    = st;
+        gain        = 1.;
+        if (gain > min_gain)
+            return gain;
+        else
+            return 0.;
+    }
+
+    /* sort in ascending order */
+    std::sort(ix_arr + st, ix_arr + end + 1, [&x](const size_t a, const size_t b){return x[a] < x[b];});
+    if (x[ix_arr[st]] == x[ix_arr[end]]) return -HUGE_VAL;
+    xmin = x[ix_arr[st]]; xmax = x[ix_arr[end]];
+
+    /* unlike the previous case for the extended model, the data here has not been centered,
+       which could make the standard deviations have poor precision. It's nevertheless not
+       necessary for this mean to have good precision, since it's only meant for centering,
+       so it can be calculated inexactly with simd instructions. */
+    real_t_ xmean = 0;
+    real_t_ cnt = 0;
+    for (size_t ix = st; ix <= end; ix++)
+    {
+        xmean += x[ix_arr[ix]];
+        cnt += w[ix_arr[ix]];
+    }
+    xmean /= cnt;
+
+    gain = find_split_std_gain_weighted(x, xmean, ix_arr, st, end, buffer_sd, criterion, min_gain, split_point, split_ix, w);
+    /* Note: a gain of -Inf signals that the data is unsplittable. Zero signals it's below the minimum. */
+    return std::fmax(0., gain);
+}
+
 /* TODO: here it should only need to look at the non-zero entries. It can then use the
    same algorithm as above, but putting an extra check to see where do the zeros fit in
    the sorted order of the non-zero entries while calculating gains and SDs, and then
@@ -761,6 +1199,28 @@ double eval_guided_crit(size_t ix_arr[], size_t st, size_t end,
                             xmin, xmax, criterion, min_gain, missing_action);
 }
 
+template <class real_t_, class sparse_ix, class mapping>
+double eval_guided_crit_weighted(size_t ix_arr[], size_t st, size_t end,
+                                 size_t col_num, real_t_ Xc[], sparse_ix Xc_ind[], sparse_ix Xc_indptr[],
+                                 double buffer_arr[], size_t buffer_pos[], bool as_relative_gain,
+                                 double &split_point, double &xmin, double &xmax,
+                                 GainCriterion criterion, double min_gain, MissingAction missing_action,
+                                 mapping w)
+{
+    todense(ix_arr, st, end,
+            col_num, Xc, Xc_ind, Xc_indptr,
+            buffer_arr);
+    std::iota(buffer_pos, buffer_pos + (end - st + 1), (size_t)0);
+    /* TODO: allocate this buffer externally */
+    std::vector<double> buffer_w(end-st+1);
+    for (size_t row = st; row <= end; row++)
+        buffer_w[row-st] = w[ix_arr[row]];
+    size_t ignored;
+    return eval_guided_crit_weighted(buffer_pos, 0, end - st, buffer_arr, buffer_arr + (end-st+1),
+                                     as_relative_gain, ignored, split_point,
+                                     xmin, xmax, criterion, min_gain, missing_action, buffer_w);
+}
+
 /* How this works:
    - For Averaged criterion, will take the expected standard deviation that would be gotten with the category counts
      if each category got assigned a real number at random ~ Unif(0,1) and the data were thus converted to
@@ -781,7 +1241,8 @@ double eval_guided_crit(size_t ix_arr[], size_t st, size_t end,
 double eval_guided_crit(size_t *restrict ix_arr, size_t st, size_t end, int *restrict x, int ncat,
                         size_t *restrict buffer_cnt, size_t *restrict buffer_pos, double *restrict buffer_prob,
                         int &chosen_cat, char *restrict split_categ, char *restrict buffer_split,
-                        GainCriterion criterion, double min_gain, bool all_perm, MissingAction missing_action, CategSplit cat_split_type)
+                        GainCriterion criterion, double min_gain, bool all_perm,
+                        MissingAction missing_action, CategSplit cat_split_type)
 {
     /* move NAs to the front if there's any, exclude them from calculations */
     if (missing_action != Fail)
@@ -1054,3 +1515,287 @@ double eval_guided_crit(size_t *restrict ix_arr, size_t st, size_t end, int *res
     else
         return best_gain;
 }
+
+
+template <class mapping>
+double eval_guided_crit_weighted(size_t *restrict ix_arr, size_t st, size_t end, int *restrict x, int ncat,
+                                 size_t *restrict buffer_pos, double *restrict buffer_prob,
+                                 int &chosen_cat, char *restrict split_categ, char *restrict buffer_split,
+                                 GainCriterion criterion, double min_gain, bool all_perm,
+                                 MissingAction missing_action, CategSplit cat_split_type,
+                                 mapping w)
+{
+    /* move NAs to the front if there's any, exclude them from calculations */
+    if (missing_action != Fail)
+        st = move_NAs_to_front(ix_arr, st, end, x);
+
+    if (st >= end) return -HUGE_VAL;
+
+    /* count categories */
+    /* TODO: allocate this buffer externally */
+    std::vector<long double> buffer_cnt(ncat, 0.);
+    for (size_t row = st; row <= end; row++)
+        buffer_cnt[x[ix_arr[row]]] += w[ix_arr[row]];
+    long double cnt = std::accumulate(buffer_cnt.begin(), buffer_cnt.end(), (long double)0);
+
+    double this_gain = -HUGE_VAL;
+    double best_gain = -HUGE_VAL;
+    std::iota(buffer_pos, buffer_pos + ncat, (size_t)0);
+    size_t st_pos = 0;
+
+    switch(cat_split_type)
+    {
+        case SingleCateg:
+        {
+            size_t ncat_present = 0;
+
+            switch(criterion)
+            {
+                case Averaged:
+                {
+                    /* move zero-counts to the beginning */
+                    size_t temp;
+                    for (int cat = 0; cat < ncat; cat++)
+                    {
+                        if (buffer_cnt[cat])
+                        {
+                            ncat_present++;
+                            buffer_prob[cat] = buffer_cnt[cat] / cnt;
+                        }
+
+                        else
+                        {
+                            temp = buffer_pos[st_pos];
+                            buffer_pos[st_pos] = buffer_pos[cat];
+                            buffer_pos[cat] = temp;
+                            st_pos++;
+                        }
+                    }
+                    
+                    if (ncat_present <= 1) return -HUGE_VAL;
+
+                    double sd_full = expected_sd_cat(buffer_prob, ncat_present, buffer_pos + st_pos);
+
+                    /* try isolating each category one at a time */
+                    for (size_t pos = st_pos; (int)pos < ncat; pos++)
+                    {
+                        this_gain = sd_gain(sd_full,
+                                            0.0,
+                                            expected_sd_cat_single(buffer_cnt.data(), buffer_prob, ncat_present, buffer_pos + st_pos, pos - st_pos, cnt)
+                                            );
+                        if (this_gain > min_gain && this_gain > best_gain)
+                        {
+                            best_gain = this_gain;
+                            chosen_cat = buffer_pos[pos];
+                        }
+                    }
+                    break;
+                }
+
+                case Pooled:
+                {
+                    /* here it will always pick the largest one */
+                    size_t ncat_present = 0;
+                    long double cnt_max = 0;
+                    for (int cat = 0; cat < ncat; cat++)
+                    {
+                        if (buffer_cnt[cat])
+                        {
+                            ncat_present++;
+                            if (cnt_max < buffer_cnt[cat])
+                            {
+                                cnt_max = buffer_cnt[cat];
+                                chosen_cat = cat;
+                            }
+                        }
+                    }
+                    
+                    if (ncat_present <= 1) return -HUGE_VAL;
+
+                    long double cnt_left = (long double)(cnt - cnt_max);
+
+                    /* TODO: think of a better way of dealing with numbers between zero and one */
+                    this_gain = (
+                                    std::fmax(1., cnt) * logl(std::fmax(1., cnt))
+                                        - std::fmax(1., cnt_left) * logl(std::fmax(1., cnt_left))
+                                        - std::fmax(1., cnt_max) * logl(std::fmax(1., cnt_max))
+                                ) / std::fmax(1., cnt);
+                    best_gain = (this_gain > min_gain)? this_gain : best_gain;
+                    break;
+                }
+            }
+            break;
+        }
+
+        case SubSet:
+        {
+            /* sort by counts */
+            std::sort(buffer_pos, buffer_pos + ncat, [&buffer_cnt](const size_t a, const size_t b){return buffer_cnt[a] < buffer_cnt[b];});
+
+            /* set split as: (1):left (0):right (-1):not_present */
+            memset(buffer_split, 0, ncat * sizeof(char));
+
+
+            switch(criterion)
+            {
+                case Averaged:
+                {
+                    /* determine first non-zero and convert to probabilities */
+                    double sd_full;
+                    for (int cat = 0; cat < ncat; cat++)
+                    {
+                        if (buffer_cnt[buffer_pos[cat]])
+                        {
+                            buffer_prob[buffer_pos[cat]] = (long double)buffer_cnt[buffer_pos[cat]] / cnt;
+                        }
+
+                        else
+                        {
+                            buffer_split[buffer_pos[cat]] = -1;
+                            st_pos++;
+                        }
+                    }
+
+                    if ((int)st_pos >= (ncat-1)) return -HUGE_VAL;
+
+                    /* calculate full SD assuming they take values randomly ~Unif(0, 1) */
+                    size_t ncat_present = (size_t)ncat - st_pos;
+                    sd_full = expected_sd_cat(buffer_prob, ncat_present, buffer_pos + st_pos);
+                    if (ncat_present >= log2ceil(SIZE_MAX)) all_perm = false;
+
+                    /* move categories one at a time */
+                    for (size_t pos = st_pos; pos < ((size_t)ncat - st_pos - 1); pos++)
+                    {
+                        buffer_split[buffer_pos[pos]] = 1;
+                        /* TODO: is this correct? */
+                        this_gain = sd_gain(sd_full,
+                                            expected_sd_cat(buffer_cnt.data(), buffer_prob, pos - st_pos + 1,       buffer_pos + st_pos),
+                                            expected_sd_cat(buffer_cnt.data(), buffer_prob, (size_t)ncat - pos - 1, buffer_pos + pos + 1)
+                                            );
+                        if (this_gain > min_gain && this_gain > best_gain)
+                        {
+                            best_gain = this_gain;
+                            memcpy(split_categ, buffer_split, ncat * sizeof(char));
+                        }
+                    }
+
+                    break;
+                }
+
+                case Pooled:
+                {
+                    long double s = 0;
+
+                    /* determine first non-zero and get base info */
+                    for (int cat = 0; cat < ncat; cat++)
+                    {
+                        if (buffer_cnt[buffer_pos[cat]])
+                        {
+                            s += (buffer_cnt[buffer_pos[cat]] <= 1)?
+                                  0 : ((long double) buffer_cnt[buffer_pos[cat]] * logl((long double)buffer_cnt[buffer_pos[cat]]));
+                        }
+
+                        else
+                        {
+                            buffer_split[buffer_pos[cat]] = -1;
+                            st_pos++;
+                        }
+                    }
+
+                    if ((int)st_pos >= (ncat-1)) return -HUGE_VAL;
+
+                    /* calculate base info */
+                    long double base_info = std::fmax(1., cnt) * logl(std::fmax(1., cnt)) - s;
+
+                    if (all_perm)
+                    {
+                        size_t cnt_left, cnt_right;
+                        double s_left, s_right;
+                        size_t ncat_present = (size_t)ncat - st_pos;
+                        size_t ncomb = pow2(ncat_present) - 1;
+                        size_t best_combin;
+
+                        for (size_t combin = 1; combin < ncomb; combin++)
+                        {
+                            cnt_left = 0; cnt_right = 0;
+                            s_left   = 0;   s_right = 0;
+                            for (size_t pos = st_pos; (int)pos < ncat; pos++)
+                            {
+                                if (extract_bit(combin, pos))
+                                {
+                                    cnt_left += buffer_cnt[buffer_pos[pos]];
+                                    s_left   += (buffer_cnt[buffer_pos[pos]] <= 1)?
+                                                 0 : ((long double) buffer_cnt[buffer_pos[pos]]
+                                                       * logl((long double) buffer_cnt[buffer_pos[pos]]));
+                                }
+
+                                else
+                                {
+                                    cnt_right += buffer_cnt[buffer_pos[pos]];
+                                    s_right   += (buffer_cnt[buffer_pos[pos]] <= 1)?
+                                                  0 : ((long double) buffer_cnt[buffer_pos[pos]]
+                                                        * logl((long double) buffer_cnt[buffer_pos[pos]]));
+                                }
+                            }
+
+                            this_gain  = categ_gain(cnt_left, cnt_right,
+                                                    s_left, s_right,
+                                                    base_info, cnt);
+
+                            if (this_gain > min_gain && this_gain > best_gain)
+                            {
+                                best_gain = this_gain;
+                                best_combin = combin;
+                            }
+
+                        }
+
+                        if (best_gain > min_gain)
+                            for (size_t pos = 0; pos < ncat_present; pos++)
+                                split_categ[buffer_pos[st_pos + pos]] = extract_bit(best_combin, pos);
+
+                    }
+
+                    else
+                    {
+                        /* try moving the categories one at a time */
+                        size_t cnt_left = 0;
+                        size_t cnt_right = end - st + 1;
+                        double s_left = 0;
+                        double s_right = s;
+
+                        for (size_t pos = st_pos; pos < (ncat - st_pos - 1); pos++)
+                        {
+                            buffer_split[buffer_pos[pos]] = 1;
+                            s_left    += (buffer_cnt[buffer_pos[pos]] <= 1)?
+                                          0 : ((long double)buffer_cnt[buffer_pos[pos]] * logl((long double)buffer_cnt[buffer_pos[pos]]));
+                            s_right   -= (buffer_cnt[buffer_pos[pos]] <= 1)?
+                                          0 : ((long double)buffer_cnt[buffer_pos[pos]] * logl((long double)buffer_cnt[buffer_pos[pos]]));
+                            cnt_left  += buffer_cnt[buffer_pos[pos]];
+                            cnt_right -= buffer_cnt[buffer_pos[pos]];
+
+                            this_gain  = categ_gain(cnt_left, cnt_right,
+                                                    s_left, s_right,
+                                                    base_info, cnt);
+
+                            if (this_gain > min_gain && this_gain > best_gain)
+                            {
+                                best_gain = this_gain;
+                                memcpy(split_categ, buffer_split, ncat * sizeof(char));
+                            }
+                        }
+                    }
+
+                    break;
+                }
+            }
+        }
+    }
+
+    if (st == (end-1)) return 0;
+
+    if (best_gain <= -HUGE_VAL && this_gain <= min_gain && this_gain > -HUGE_VAL)
+        return 0;
+    else
+        return best_gain;
+}
